
#include "JViewWindow.h"
#include "JViewView.h"
#include "FileEntry.h"
#include "JVConstants.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>

extern "C" {
#define HAVE_PROTOTYPES
#include "jpeglib.h"
}

#define IMAGE_TOP_Y		24.0

JViewWindow::JViewWindow(BRect frame, BList *files)
				: BWindow(frame, "JView", B_TITLED_WINDOW, 
				B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	FileEntry *entry;
	BRect rect;
	
	// Save list of files to be shown
	entry_list = files;
	
	// Start converter thread
	conv_thread = spawn_thread(entry_func, "Converter",
								B_NORMAL_PRIORITY, this);
	if (conv_thread < B_NO_ERROR) {
		fprintf(stderr, "Can't create converter thread.\n");
		Quit();
	}
	if (resume_thread(conv_thread) != B_NO_ERROR) {
		fprintf(stderr, "Error starting converter thread.\n");
		Quit();
	}
	current = 0;
	conv_index = 0;
	entry = (FileEntry *) entry_list->FirstItem();
	send_data(conv_thread, START, &entry, sizeof(entry));
	
	// Create and add GUI components
	rect.Set(0.0, 0.0, 49.0, 17.0);
	next_button = new BButton(rect, "next", "Next",
								new BMessage(NEXT), B_FOLLOW_NONE);
	next_button->SetEnabled(FALSE);
	next_button->SetFontSize(9);
	AddChild(next_button);
	
	rect.Set(51.0, 0.0, 100.0, 17.0);
	del = new BButton(rect, "Delete", "Delete",
								new BMessage(DELETE), B_FOLLOW_NONE);
	del->SetFontSize(9);
	AddChild(del);
	
	rect.Set(109.0, 0.0, 199.0, 18.0);	
	label = new BStringView(rect, "label", "", B_FOLLOW_LEFT_RIGHT);
	AddChild(label);
}

bool
JViewWindow::QuitRequested()
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return(TRUE);
}


//
// Message handler
//
void
JViewWindow::MessageReceived(BMessage *message) {
	FileEntry *entry;
	
	entry = (FileEntry *) entry_list->ItemAt(current);
	switch (message->what) {
	case DELETE:
//		printf("Trying to delete %s\n", entry->Name());
		if (remove(entry->Name()))
			printf("errno=%d\n", errno);
		if (current == (entry_list->CountItems() - 1)) {
			be_app->PostMessage(B_QUIT_REQUESTED);
			Quit();
		}
	case NEXT:
		// Delete current image
		delete entry->Image();
		entry->Image(NULL);
		entry->Lines(0);
		entry->MaxLines(0);
		current++;
		if (current < conv_index) {
			// We aren't waiting on converter to complete
			Draw(TRUE);

			if (conv_index < entry_list->CountItems()) {
				// Start converting next image
				entry = (FileEntry *) entry_list->ItemAt(conv_index);
				send_data(conv_thread, START, &entry, sizeof(entry));
			}
			else {	// conv_index == CountItems()
				Lock();
				next_button->SetEnabled(FALSE);
				Unlock();
			}
		}
		else {
			// Waiting on the converter
			Lock();
			next_button->SetEnabled(FALSE);
			del->SetEnabled(FALSE);
			label->SetText("Converting ...");
			Unlock();
		}
		break;
	case FINISHED:
		if (current == conv_index) {
			// Have been waiting on converter
			Draw(TRUE);
			if (current < (entry_list->CountItems()-1)) {
				Lock();
				next_button->SetEnabled(TRUE);
				Unlock();

				// Start converting next image
				conv_index++;
				entry = (FileEntry *) entry_list->ItemAt(conv_index);
				send_data(conv_thread, START, &entry, sizeof(entry));
			}
			else {
				// Shut down converter thread
				send_data(conv_thread, ABORT, NULL, 0);
			}
		}
		else
			conv_index++;
		break;
	default:
		BWindow::MessageReceived(message);
	}
}



//
// Function to intercept key down events
//
bool
JViewWindow::FilterKeyDown(ulong *aChar, BView **) {
	if (*aChar == '\n')
		return(TRUE);
	if (*aChar == ' ') {
		if (next_button->IsEnabled())
			PostMessage(NEXT);
		else if (current == (entry_list->CountItems() - 1)) {
			be_app->PostMessage(B_QUIT_REQUESTED);
			Quit();
		}
	}
	return(FALSE);
}


//
// Function to redraw the window
//
void
JViewWindow::Draw(bool resize) {
	FileEntry *entry;
	BBitmap *BM;
	JViewView *view;
	BRect rect;
	static char lbuf[64];
	const char *fname;
	
	entry = (FileEntry *) entry_list->ItemAt(current);
	BM = entry->Image();
			
	if ((BM != NULL) && (entry->Lines() == entry->MaxLines())) {
		Lock();
		view = (JViewView *) FindView("JViewView");
		if (resize) {
			rect = BM->Bounds();
			ResizeTo(rect.Width()-1.0, rect.Height()+IMAGE_TOP_Y-1.0);
			
			// Update label
			fname = strrchr(entry->Name(), '/');
			if (fname == NULL)
				fname = entry->Name();
			else
				fname++;
			sprintf(lbuf, "%s (%dx%d)", fname, rect.IntegerWidth(),
					rect.IntegerHeight());
			label->SetText(lbuf);
		}
		view->DrawBitmap(BM, BPoint(0.0, IMAGE_TOP_Y));
		del->SetEnabled(TRUE);		
		Unlock();
	}
}


//
// Literal entry function
//
long
JViewWindow::entry_func(void *arg) {
	JViewWindow *window;
	
	window = (JViewWindow *) arg;
	return(window->converter());
}


//
// Actual convertor function (body of converter thread
//
long
JViewWindow::converter() {
	FileEntry *entry;
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	FILE *infile;
	thread_id sender;
	BRect rect;
	BBitmap *BM;
	JSAMPARRAY buffer;
	int row_size;
	int lines_read;
	long row_offset;
		
	// Allocate buffer for row pointers
	buffer = new JSAMPROW;
	
	while (receive_data(&sender, (void *) &entry, sizeof(entry))
			== START) {
		
		// Allocate a decompression object and initialize
		cinfo.err = jpeg_std_error(&jerr);
		jpeg_create_decompress(&cinfo);
		
		// Set up input stream
		if (!(infile = fopen(entry->Name(), "rb"))) {
			// error opening file. Skip it.
			fprintf(stderr, "Can't open file '%s'\n", entry->Name());
			be_app->PostMessage(FINISHED);
			jpeg_destroy_decompress(&cinfo);
			continue;
		}
		jpeg_stdio_src(&cinfo, infile);
		
		// Read header and allocate an appropriately sized
		// bitmap.
		jpeg_read_header(&cinfo, TRUE);
		
		rect.Set(0.0, 0.0, cinfo.image_width, cinfo.image_height);
		BM = new BBitmap(rect, B_RGB_32_BIT);
		entry->Lines(0);
		entry->MaxLines(cinfo.image_height);
		entry->Image(BM);
		row_size = cinfo.num_components * cinfo.image_width;
		buffer[0] = new JSAMPLE[row_size];
		
		// Start decompression
		jpeg_start_decompress(&cinfo);
		
		row_offset = 0;
		while (cinfo.output_scanline < cinfo.output_height) {
			lines_read = jpeg_read_scanlines(&cinfo, buffer, 1);
			
			// Report progress
			entry->Lines(cinfo.output_scanline);
			
			// Copy to bitmap
			if (cinfo.output_components ==1) {
				// Grayscale
				BM->SetBits(buffer[0], row_size, row_offset,
							B_GRAYSCALE_8_BIT);
			}
			else {
				// Color
				BM->SetBits(buffer[0], row_size, row_offset,
							B_RGB_32_BIT);
			}
			
			row_offset += BM->BytesPerRow();
			
			if (has_data(find_thread(NULL))) {
				// Incoming message. Abort current image.
				printf("Error! Abort occurred\n");
				delete BM;
				entry->Image(NULL);
				entry->Lines(0);
				entry->MaxLines(0);
				goto aborted;
			}
		}
		
		jpeg_finish_decompress(&cinfo);
		
aborted:
		jpeg_destroy_decompress(&cinfo);
		fclose(infile);
		
		// Signal completion
		be_app->PostMessage(FINISHED);
	}
	
	delete buffer;
	return(0);
}
